# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.alert_page_public import AlertPagePublic
from ..types.breakdown_config_public import BreakdownConfigPublic
from ..types.dashboard_page_public import DashboardPagePublic
from ..types.dataset_page_public import DatasetPagePublic
from ..types.experiment_page_public import ExperimentPagePublic
from ..types.feedback_score_names import FeedbackScoreNames
from ..types.kpi_card_response import KpiCardResponse
from ..types.optimization_page_public import OptimizationPagePublic
from ..types.project_detailed import ProjectDetailed
from ..types.project_metric_response_public import ProjectMetricResponsePublic
from ..types.project_page_public import ProjectPagePublic
from ..types.project_public import ProjectPublic
from ..types.project_stats_summary import ProjectStatsSummary
from ..types.prompt_page_public import PromptPagePublic
from ..types.span_filter_public import SpanFilterPublic
from ..types.token_usage_names import TokenUsageNames
from ..types.trace_filter_public import TraceFilterPublic
from ..types.trace_thread_filter_public import TraceThreadFilterPublic
from .raw_client import AsyncRawProjectsClient, RawProjectsClient
from .types.kpi_card_request_entity_type import KpiCardRequestEntityType
from .types.project_metric_request_public_interval import ProjectMetricRequestPublicInterval
from .types.project_metric_request_public_metric_type import ProjectMetricRequestPublicMetricType
from .types.project_update_visibility import ProjectUpdateVisibility
from .types.project_write_visibility import ProjectWriteVisibility

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ProjectsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawProjectsClient
        """
        return self._raw_client

    def find_alerts_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AlertPagePublic:
        """
        Find alerts scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPagePublic
            Alerts page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_alerts_by_project(project_id='projectId', )
        """
        _response = self._raw_client.find_alerts_by_project(
            project_id, page=page, size=size, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    def find_dashboards_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPagePublic:
        """
        Find dashboards scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPagePublic
            Dashboard page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_dashboards_by_project(project_id='projectId', )
        """
        _response = self._raw_client.find_dashboards_by_project(
            project_id,
            page=page,
            size=size,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def find_datasets_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        with_experiments_only: typing.Optional[bool] = None,
        with_optimizations_only: typing.Optional[bool] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DatasetPagePublic:
        """
        Find datasets scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        with_experiments_only : typing.Optional[bool]

        with_optimizations_only : typing.Optional[bool]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DatasetPagePublic
            Dataset page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_datasets_by_project(project_id='projectId', )
        """
        _response = self._raw_client.find_datasets_by_project(
            project_id,
            page=page,
            size=size,
            with_experiments_only=with_experiments_only,
            with_optimizations_only=with_optimizations_only,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def find_experiments_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        optimization_id: typing.Optional[str] = None,
        types: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        experiment_ids: typing.Optional[str] = None,
        force_sorting: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExperimentPagePublic:
        """
        Find experiments scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        optimization_id : typing.Optional[str]

        types : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        experiment_ids : typing.Optional[str]

        force_sorting : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPagePublic
            Experiments page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_experiments_by_project(project_id='projectId', )
        """
        _response = self._raw_client.find_experiments_by_project(
            project_id,
            page=page,
            size=size,
            dataset_id=dataset_id,
            optimization_id=optimization_id,
            types=types,
            name=name,
            dataset_deleted=dataset_deleted,
            sorting=sorting,
            filters=filters,
            experiment_ids=experiment_ids,
            force_sorting=force_sorting,
            request_options=request_options,
        )
        return _response.data

    def find_optimizations_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        dataset_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizationPagePublic:
        """
        Find optimizations scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        dataset_name : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPagePublic
            Optimizations page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_optimizations_by_project(project_id='projectId', )
        """
        _response = self._raw_client.find_optimizations_by_project(
            project_id,
            page=page,
            size=size,
            dataset_id=dataset_id,
            dataset_name=dataset_name,
            name=name,
            dataset_deleted=dataset_deleted,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def get_prompts_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPagePublic:
        """
        Get prompts scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPagePublic
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_prompts_by_project(project_id='projectId', )
        """
        _response = self._raw_client.get_prompts_by_project(
            project_id,
            page=page,
            size=size,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectPagePublic:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPagePublic
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_projects()
        """
        _response = self._raw_client.find_projects(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.create_project(name='name', )
        """
        _response = self._raw_client.create_project(
            name=name, visibility=visibility, description=description, request_options=request_options
        )
        return _response.data

    def get_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> ProjectPublic:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPublic
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_by_id(id='id', )
        """
        _response = self._raw_client.get_project_by_id(id, request_options=request_options)
        return _response.data

    def delete_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.delete_project_by_id(id='id', )
        """
        _response = self._raw_client.delete_project_by_id(id, request_options=request_options)
        return _response.data

    def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.update_project(id='id', )
        """
        _response = self._raw_client.update_project(
            id, name=name, description=description, visibility=visibility, request_options=request_options
        )
        return _response.data

    def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.delete_projects_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_projects_batch(ids=ids, request_options=request_options)
        return _response.data

    def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackScoreNames:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNames
            Feedback Scores resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_feedback_score_names_by_project_ids()
        """
        _response = self._raw_client.find_feedback_score_names_by_project_ids(
            project_ids=project_ids, request_options=request_options
        )
        return _response.data

    def find_token_usage_names(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> TokenUsageNames:
        """
        Find Token Usage names

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        TokenUsageNames
            Token Usage names resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_token_usage_names(id='id', )
        """
        _response = self._raw_client.find_token_usage_names(id, request_options=request_options)
        return _response.data

    def get_project_kpi_cards(
        self,
        id: str,
        *,
        entity_type: KpiCardRequestEntityType,
        interval_start: dt.datetime,
        interval_end: dt.datetime,
        filters: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> KpiCardResponse:
        """
        Gets KPI card metrics for a project

        Parameters
        ----------
        id : str

        entity_type : KpiCardRequestEntityType

        interval_start : dt.datetime

        interval_end : dt.datetime

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        KpiCardResponse
            KPI Card Metrics

        Examples
        --------
        from Opik import OpikApi
        import datetime
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_kpi_cards(id='id', entity_type="traces", interval_start=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), interval_end=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )
        """
        _response = self._raw_client.get_project_kpi_cards(
            id,
            entity_type=entity_type,
            interval_start=interval_start,
            interval_end=interval_end,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        span_filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        trace_filters: typing.Optional[typing.Sequence[TraceFilterPublic]] = OMIT,
        thread_filters: typing.Optional[typing.Sequence[TraceThreadFilterPublic]] = OMIT,
        breakdown: typing.Optional[BreakdownConfigPublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectMetricResponsePublic:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        span_filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        trace_filters : typing.Optional[typing.Sequence[TraceFilterPublic]]

        thread_filters : typing.Optional[typing.Sequence[TraceThreadFilterPublic]]

        breakdown : typing.Optional[BreakdownConfigPublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectMetricResponsePublic
            Project Metrics

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_metrics(id='id', )
        """
        _response = self._raw_client.get_project_metrics(
            id,
            metric_type=metric_type,
            interval=interval,
            interval_start=interval_start,
            interval_end=interval_end,
            span_filters=span_filters,
            trace_filters=trace_filters,
            thread_filters=thread_filters,
            breakdown=breakdown,
            request_options=request_options,
        )
        return _response.data

    def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsSummary:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsSummary
            Project Stats

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_stats()
        """
        _response = self._raw_client.get_project_stats(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectDetailed:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectDetailed
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.retrieve_project(name='name', )
        """
        _response = self._raw_client.retrieve_project(name=name, request_options=request_options)
        return _response.data


class AsyncProjectsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawProjectsClient
        """
        return self._raw_client

    async def find_alerts_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AlertPagePublic:
        """
        Find alerts scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AlertPagePublic
            Alerts page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_alerts_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_alerts_by_project(
            project_id, page=page, size=size, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    async def find_dashboards_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPagePublic:
        """
        Find dashboards scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPagePublic
            Dashboard page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_dashboards_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_dashboards_by_project(
            project_id,
            page=page,
            size=size,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def find_datasets_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        with_experiments_only: typing.Optional[bool] = None,
        with_optimizations_only: typing.Optional[bool] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DatasetPagePublic:
        """
        Find datasets scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        with_experiments_only : typing.Optional[bool]

        with_optimizations_only : typing.Optional[bool]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DatasetPagePublic
            Dataset page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_datasets_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_datasets_by_project(
            project_id,
            page=page,
            size=size,
            with_experiments_only=with_experiments_only,
            with_optimizations_only=with_optimizations_only,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def find_experiments_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        optimization_id: typing.Optional[str] = None,
        types: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        experiment_ids: typing.Optional[str] = None,
        force_sorting: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExperimentPagePublic:
        """
        Find experiments scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        optimization_id : typing.Optional[str]

        types : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        experiment_ids : typing.Optional[str]

        force_sorting : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPagePublic
            Experiments page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_experiments_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_experiments_by_project(
            project_id,
            page=page,
            size=size,
            dataset_id=dataset_id,
            optimization_id=optimization_id,
            types=types,
            name=name,
            dataset_deleted=dataset_deleted,
            sorting=sorting,
            filters=filters,
            experiment_ids=experiment_ids,
            force_sorting=force_sorting,
            request_options=request_options,
        )
        return _response.data

    async def find_optimizations_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        dataset_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizationPagePublic:
        """
        Find optimizations scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        dataset_name : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPagePublic
            Optimizations page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_optimizations_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_optimizations_by_project(
            project_id,
            page=page,
            size=size,
            dataset_id=dataset_id,
            dataset_name=dataset_name,
            name=name,
            dataset_deleted=dataset_deleted,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def get_prompts_by_project(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPagePublic:
        """
        Get prompts scoped to a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPagePublic
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_prompts_by_project(project_id='projectId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompts_by_project(
            project_id,
            page=page,
            size=size,
            name=name,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectPagePublic:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPagePublic
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_projects()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_projects(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    async def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.create_project(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_project(
            name=name, visibility=visibility, description=description, request_options=request_options
        )
        return _response.data

    async def get_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectPublic:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPublic
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_by_id(id, request_options=request_options)
        return _response.data

    async def delete_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.delete_project_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_project_by_id(id, request_options=request_options)
        return _response.data

    async def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.update_project(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_project(
            id, name=name, description=description, visibility=visibility, request_options=request_options
        )
        return _response.data

    async def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.delete_projects_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_projects_batch(ids=ids, request_options=request_options)
        return _response.data

    async def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackScoreNames:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNames
            Feedback Scores resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_feedback_score_names_by_project_ids()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_feedback_score_names_by_project_ids(
            project_ids=project_ids, request_options=request_options
        )
        return _response.data

    async def find_token_usage_names(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> TokenUsageNames:
        """
        Find Token Usage names

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        TokenUsageNames
            Token Usage names resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_token_usage_names(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.find_token_usage_names(id, request_options=request_options)
        return _response.data

    async def get_project_kpi_cards(
        self,
        id: str,
        *,
        entity_type: KpiCardRequestEntityType,
        interval_start: dt.datetime,
        interval_end: dt.datetime,
        filters: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> KpiCardResponse:
        """
        Gets KPI card metrics for a project

        Parameters
        ----------
        id : str

        entity_type : KpiCardRequestEntityType

        interval_start : dt.datetime

        interval_end : dt.datetime

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        KpiCardResponse
            KPI Card Metrics

        Examples
        --------
        from Opik import AsyncOpikApi
        import datetime
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_kpi_cards(id='id', entity_type="traces", interval_start=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), interval_end=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_kpi_cards(
            id,
            entity_type=entity_type,
            interval_start=interval_start,
            interval_end=interval_end,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        span_filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        trace_filters: typing.Optional[typing.Sequence[TraceFilterPublic]] = OMIT,
        thread_filters: typing.Optional[typing.Sequence[TraceThreadFilterPublic]] = OMIT,
        breakdown: typing.Optional[BreakdownConfigPublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectMetricResponsePublic:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        span_filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        trace_filters : typing.Optional[typing.Sequence[TraceFilterPublic]]

        thread_filters : typing.Optional[typing.Sequence[TraceThreadFilterPublic]]

        breakdown : typing.Optional[BreakdownConfigPublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectMetricResponsePublic
            Project Metrics

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_metrics(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_metrics(
            id,
            metric_type=metric_type,
            interval=interval,
            interval_start=interval_start,
            interval_end=interval_end,
            span_filters=span_filters,
            trace_filters=trace_filters,
            thread_filters=thread_filters,
            breakdown=breakdown,
            request_options=request_options,
        )
        return _response.data

    async def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsSummary:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsSummary
            Project Stats

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_stats()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_stats(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    async def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectDetailed:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectDetailed
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.retrieve_project(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve_project(name=name, request_options=request_options)
        return _response.data



"""
Decorator for OpenAI TTS API calls (audio.speech.create).

Output type: openai.types.audio.speech.HttpxBinaryResponseContent

This decorator tracks synchronous TTS API calls, logging:
- Input parameters (model, voice, input text, response_format, speed)
- Usage (input_characters count mapped to prompt_tokens/total_tokens)
- Model and provider metadata
"""

import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from typing_extensions import override

import opik.dict_utils as dict_utils
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator

LOGGER = logging.getLogger(__name__)

# Input parameters to log for TTS creation
TTS_CREATE_KWARGS_KEYS_TO_LOG_AS_INPUTS = [
    "input",
    "voice",
    "response_format",
    "speed",
    "instructions",
]

# Response metadata keys to extract
TTS_RESPONSE_KEYS_TO_LOG_AS_OUTPUT = [
    "content_type",
    "content_length",
]


class TTSCreateTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    Decorator for tracking OpenAI TTS API calls (audio.speech.create).

    Handles: audio.speech.create (synchronous, non-streaming)

    The TTS API returns binary audio content (HttpxBinaryResponseContent).
    We track input parameters and character count usage, but do NOT buffer
    the audio content to avoid excessive memory usage.
    """

    def __init__(self, provider: str) -> None:
        super().__init__()
        self.provider = provider

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        assert kwargs is not None, (
            "Expected kwargs to be not None in audio.speech.create(**kwargs)"
        )

        name = track_options.name if track_options.name is not None else func.__name__

        metadata = track_options.metadata if track_options.metadata is not None else {}

        input_data, new_metadata = dict_utils.split_dict_by_keys(
            kwargs, keys=TTS_CREATE_KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata.update(
            {
                "created_from": "openai",
                "type": "openai_tts",
            }
        )

        tags = ["openai", "tts"]
        model = kwargs.get("model", "tts-1")

        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input_data,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
            model=model,
            provider=self.provider,
        )

        return result

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        usage = None
        output_data: Optional[Dict[str, Any]] = None
        metadata: Dict[str, Any] = {}

        # Calculate character-based usage from the input text
        # The input text is stored in the span's input data
        input_text = ""
        if current_span_data.input is not None:
            input_text = current_span_data.input.get("input", "")

        input_characters = len(input_text) if isinstance(input_text, str) else 0

        if input_characters > 0:
            # Map character count to token fields for backend compatibility
            # The backend will use input_characters for audio_speech cost calculation
            usage = {
                "completion_tokens": 0,
                "prompt_tokens": input_characters,
                "total_tokens": input_characters,
                "input_characters": input_characters,
            }
            metadata["input_characters"] = input_characters

        # Extract what we can from the response without consuming the binary content
        if output is not None:
            try:
                # HttpxBinaryResponseContent has a response attribute
                if hasattr(output, "response"):
                    response = output.response
                    content_type = response.headers.get("content-type", "audio/mpeg")
                    content_length = response.headers.get("content-length")
                    output_data = {
                        "content_type": content_type,
                    }
                    if content_length is not None:
                        output_data["content_length"] = int(content_length)
                        metadata["audio_bytes"] = int(content_length)
                else:
                    output_data = {"type": "audio_binary_content"}
            except Exception:
                LOGGER.debug("Failed to extract TTS response metadata", exc_info=True)
                output_data = {"type": "audio_binary_content"}

        return arguments_helpers.EndSpanParameters(
            output=output_data,
            usage=usage,
            metadata=metadata,
            model=current_span_data.model,
            provider=self.provider,
        )

    @override
    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Optional[Any]:
        NOT_A_STREAM = None
        return NOT_A_STREAM



import logging
from abc import ABC, abstractmethod
from typing import (
    Optional,
    List,
    Literal,
    Union,
)

from opik.rest_api import client as rest_api_client
from opik.rest_api.types import trace_public, trace_thread
from opik.message_processing.batching import sequence_splitter
from opik.api_objects.trace import trace_client
from opik.api_objects.rest_helpers import ensure_rest_api_call_respecting_rate_limit
from opik.api_objects import constants
import opik.exceptions as exceptions

LOGGER = logging.getLogger(__name__)

TraceType = Union[trace_client.Trace, trace_public.TracePublic]


class BaseAnnotationQueue(ABC):
    """
    Base class for annotation queue objects.

    This class provides the common functionality shared between
    TracesAnnotationQueue and ThreadsAnnotationQueue.

    This object should not be created directly, instead use the appropriate
    create/get methods on opik.Opik client.
    """

    def __init__(
        self,
        id: str,
        name: str,
        project_id: str,
        rest_client: rest_api_client.OpikApi,
        description: Optional[str] = None,
        instructions: Optional[str] = None,
        comments_enabled: Optional[bool] = None,
        feedback_definition_names: Optional[List[str]] = None,
        items_count: Optional[int] = None,
    ) -> None:
        self._id = id
        self._name = name
        self._description = description
        self._instructions = instructions
        self._project_id = project_id
        self._comments_enabled = comments_enabled
        self._feedback_definition_names = feedback_definition_names
        self._items_count = items_count
        self._rest_client = rest_client

    @property
    def id(self) -> str:
        """The id of the annotation queue."""
        return self._id

    @property
    def name(self) -> str:
        """The name of the annotation queue."""
        return self._name

    @property
    def description(self) -> Optional[str]:
        """The description of the annotation queue."""
        return self._description

    @property
    def instructions(self) -> Optional[str]:
        """The instructions for reviewers."""
        return self._instructions

    @property
    def project_id(self) -> str:
        """The project ID associated with this annotation queue."""
        return self._project_id

    @property
    def comments_enabled(self) -> Optional[bool]:
        """Whether comments are enabled for this queue."""
        return self._comments_enabled

    @property
    def feedback_definition_names(self) -> Optional[List[str]]:
        """The feedback definition names associated with this queue."""
        return self._feedback_definition_names

    @property
    @abstractmethod
    def scope(self) -> str:
        """The scope of the annotation queue ('trace' or 'thread')."""
        pass

    @property
    def items_count(self) -> Optional[int]:
        """
        The total number of items in the queue.

        If the count is not cached locally, it will be fetched from the backend.
        """
        if self._items_count is None:
            queue_info = self._rest_client.annotation_queues.get_annotation_queue_by_id(
                self._id
            )
            self._items_count = queue_info.items_count
        return self._items_count

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        instructions: Optional[str] = None,
        comments_enabled: Optional[bool] = None,
        feedback_definition_names: Optional[List[str]] = None,
    ) -> None:
        """
        Update the annotation queue properties.

        Args:
            name: New name for the queue.
            description: New description for the queue.
            instructions: New instructions for reviewers.
            comments_enabled: Whether to enable comments.
            feedback_definition_names: List of feedback definition names.
        """
        self._rest_client.annotation_queues.update_annotation_queue(
            id=self._id,
            name=name,
            description=description,
            instructions=instructions,
            comments_enabled=comments_enabled,
            feedback_definition_names=feedback_definition_names,
        )

        if name is not None:
            self._name = name
        if description is not None:
            self._description = description
        if instructions is not None:
            self._instructions = instructions
        if comments_enabled is not None:
            self._comments_enabled = comments_enabled
        if feedback_definition_names is not None:
            self._feedback_definition_names = feedback_definition_names

    def delete(self) -> None:
        """
        Delete this annotation queue.
        """
        self._rest_client.annotation_queues.delete_annotation_queue_batch(
            ids=[self._id]
        )

    def _add_items_batch_with_retry(self, ids: List[str]) -> None:
        """Add a batch of items with automatic retry on rate limit errors."""
        ensure_rest_api_call_respecting_rate_limit(
            lambda: self._rest_client.annotation_queues.add_items_to_annotation_queue(
                id=self._id, ids=ids
            )
        )
        LOGGER.debug("Successfully added %d items to annotation queue", len(ids))

    def _remove_items_batch_with_retry(self, ids: List[str]) -> None:
        """Remove a batch of items with automatic retry on rate limit errors."""
        ensure_rest_api_call_respecting_rate_limit(
            lambda: self._rest_client.annotation_queues.remove_items_from_annotation_queue(
                id=self._id, ids=ids
            )
        )
        LOGGER.debug("Successfully removed %d items from annotation queue", len(ids))


class TracesAnnotationQueue(BaseAnnotationQueue):
    """
    An annotation queue for traces.

    This queue is used to collect traces for human annotation workflows.

    This object should not be created directly, instead use
    :meth:`opik.Opik.create_traces_annotation_queue` or :meth:`opik.Opik.get_traces_annotation_queue`.
    """

    SCOPE: Literal["trace"] = "trace"

    @property
    def scope(self) -> str:
        """The scope of the annotation queue."""
        return self.SCOPE

    def _extract_trace_ids(
        self,
        traces: List[TraceType],
    ) -> List[str]:
        """Extract IDs from trace objects."""
        ids: List[str] = []
        for trace in traces:
            if trace.id is None:
                raise exceptions.OpikException("Trace object has no id")
            ids.append(trace.id)

        return ids

    def add_traces(
        self,
        traces: List[TraceType],
    ) -> None:
        """
        Add trace objects to the annotation queue.

        Args:
            traces: A list of traces to add. For a single trace, wrap it in a list: [trace].
                Accepts Trace objects (from opik_client.trace()) or TracePublic objects
                (from search_traces()).

        Raises:
            OpikException: If any trace object has no id.
        """
        ids = self._extract_trace_ids(traces)
        if not ids:
            return

        batches = sequence_splitter.split_into_batches(
            ids, max_length=constants.ANNOTATION_QUEUE_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Adding %d traces to annotation queue", len(batch))
            self._add_items_batch_with_retry(batch)

        self._items_count = None

    def remove_traces(
        self,
        traces: List[TraceType],
    ) -> None:
        """
        Remove trace objects from the annotation queue.

        Args:
            traces: A list of traces to remove. For a single trace, wrap it in a list: [trace].
                Accepts Trace objects (from opik_client.trace()) or TracePublic objects
                (from search_traces()).

        Raises:
            OpikException: If any trace object has no id.
        """
        ids = self._extract_trace_ids(traces)
        if not ids:
            return

        batches = sequence_splitter.split_into_batches(
            ids, max_length=constants.ANNOTATION_QUEUE_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Removing %d traces from annotation queue", len(batch))
            self._remove_items_batch_with_retry(batch)

        self._items_count = None


class ThreadsAnnotationQueue(BaseAnnotationQueue):
    """
    An annotation queue for threads.

    This queue is used to collect threads for human annotation workflows.

    This object should not be created directly, instead use
    :meth:`opik.Opik.create_threads_annotation_queue` or :meth:`opik.Opik.get_threads_annotation_queue`.
    """

    SCOPE: Literal["thread"] = "thread"

    @property
    def scope(self) -> str:
        """The scope of the annotation queue."""
        return self.SCOPE

    def _extract_thread_ids(
        self,
        threads: List[trace_thread.TraceThread],
    ) -> List[str]:
        """Extract thread_model_id from TraceThread objects."""
        ids: List[str] = []
        for thread in threads:
            if thread.thread_model_id is None:
                raise exceptions.OpikException(
                    "TraceThread object has no thread_model_id"
                )
            ids.append(thread.thread_model_id)

        return ids

    def add_threads(
        self,
        threads: List[trace_thread.TraceThread],
    ) -> None:
        """
        Add thread objects to the annotation queue.

        Args:
            threads: A list of TraceThread objects to add (from search_threads()).
                For a single thread, wrap it in a list: [thread].

        Raises:
            OpikException: If any thread object has no thread_model_id.
        """
        ids = self._extract_thread_ids(threads)
        if not ids:
            return

        batches = sequence_splitter.split_into_batches(
            ids, max_length=constants.ANNOTATION_QUEUE_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Adding %d threads to annotation queue", len(batch))
            self._add_items_batch_with_retry(batch)

        self._items_count = None

    def remove_threads(
        self,
        threads: List[trace_thread.TraceThread],
    ) -> None:
        """
        Remove thread objects from the annotation queue.

        Args:
            threads: A list of TraceThread objects to remove (from search_threads()).
                For a single thread, wrap it in a list: [thread].

        Raises:
            OpikException: If any thread object has no thread_model_id.
        """
        ids = self._extract_thread_ids(threads)
        if not ids:
            return

        batches = sequence_splitter.split_into_batches(
            ids, max_length=constants.ANNOTATION_QUEUE_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Removing %d threads from annotation queue", len(batch))
            self._remove_items_batch_with_retry(batch)

        self._items_count = None



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.forbidden_error import ForbiddenError
from ..errors.not_found_error import NotFoundError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.error_message import ErrorMessage
from ..types.provider_api_key_page_public import ProviderApiKeyPagePublic
from ..types.provider_api_key_public import ProviderApiKeyPublic
from .types.provider_api_key_write_provider import ProviderApiKeyWriteProvider

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawLlmProviderKeyClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def delete_llm_provider_api_keys_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete LLM Provider's ApiKeys batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def find_llm_provider_keys(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProviderApiKeyPagePublic]:
        """
        Find LLM Provider's ApiKeys

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProviderApiKeyPagePublic]
            LLMProviderApiKey resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPagePublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def store_llm_provider_api_key(
        self,
        *,
        provider: ProviderApiKeyWriteProvider,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Store LLM Provider's ApiKey

        Parameters
        ----------
        provider : ProviderApiKeyWriteProvider

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - required for custom LLM and Bedrock providers to uniquely identify them (e.g., 'ollama', 'vllm', 'Bedrock us-east-1'). Must not be blank for custom and Bedrock providers. Should not be set for standard providers (OpenAI, Anthropic, etc.). This requirement is conditional and validation is enforced programmatically.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="POST",
            json={
                "provider": provider,
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_llm_provider_api_key_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProviderApiKeyPublic]:
        """
        Get LLM Provider's ApiKey by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProviderApiKeyPublic]
            LLMProviderApiKey resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_llm_provider_api_key(
        self,
        id: str,
        *,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update LLM Provider's ApiKey

        Parameters
        ----------
        id : str

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - can be set to migrate legacy custom LLM or Bedrock providers to the new multi-provider format. Once set, it cannot be changed. Should only be set for custom LLM and Bedrock providers.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawLlmProviderKeyClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def delete_llm_provider_api_keys_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete LLM Provider's ApiKeys batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def find_llm_provider_keys(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProviderApiKeyPagePublic]:
        """
        Find LLM Provider's ApiKeys

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProviderApiKeyPagePublic]
            LLMProviderApiKey resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPagePublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def store_llm_provider_api_key(
        self,
        *,
        provider: ProviderApiKeyWriteProvider,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Store LLM Provider's ApiKey

        Parameters
        ----------
        provider : ProviderApiKeyWriteProvider

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - required for custom LLM and Bedrock providers to uniquely identify them (e.g., 'ollama', 'vllm', 'Bedrock us-east-1'). Must not be blank for custom and Bedrock providers. Should not be set for standard providers (OpenAI, Anthropic, etc.). This requirement is conditional and validation is enforced programmatically.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="POST",
            json={
                "provider": provider,
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_llm_provider_api_key_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProviderApiKeyPublic]:
        """
        Get LLM Provider's ApiKey by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProviderApiKeyPublic]
            LLMProviderApiKey resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_llm_provider_api_key(
        self,
        id: str,
        *,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update LLM Provider's ApiKey

        Parameters
        ----------
        id : str

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - can be set to migrate legacy custom LLM or Bedrock providers to the new multi-provider format. Once set, it cannot be changed. Should only be set for custom LLM and Bedrock providers.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import DatasetUpdateVisibility, DatasetWriteType, DatasetWriteVisibility

__all__ = ["DatasetUpdateVisibility", "DatasetWriteType", "DatasetWriteVisibility"]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .json_list_string_write import JsonListStringWrite
from .optimization_studio_config_write import OptimizationStudioConfigWrite
from .optimization_write_status import OptimizationWriteStatus


class OptimizationWrite(UniversalBaseModel):
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None
    dataset_name: str
    project_name: typing.Optional[str] = pydantic.Field(default=None)
    """
    Project name. Creates project if it doesn't exist. Ignored when project_id is provided.
    """

    project_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Project ID. Takes precedence over project_name when both are provided.
    """

    objective_name: str
    status: OptimizationWriteStatus
    metadata: typing.Optional[JsonListStringWrite] = None
    studio_config: typing.Optional[OptimizationStudioConfigWrite] = None
    last_updated_at: typing.Optional[dt.datetime] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import importlib.metadata


def _get_package_version() -> str:
    try:
        return importlib.metadata.version("opik")
    except importlib.metadata.PackageNotFoundError:
        return "Please install opik with `pip install opik`"


VERSION = _get_package_version()



import datetime
import logging
from typing import Optional, Dict, Any, List, TypeVar, Type, Union, Literal

import opik.llm_usage as llm_usage
from . import opik_query_language, validation_helpers, constants
from .. import config, datetime_helpers, logging_messages, id_helpers
from ..message_processing import messages
from ..rest_api.types import (
    span_filter_public,
    trace_filter_public,
    trace_thread_filter,
)
from ..types import BatchFeedbackScoreDict

# Re-export for backward compatibility
generate_id = id_helpers.generate_id

LOGGER = logging.getLogger(__name__)


FilterParsedItemT = TypeVar(
    "FilterParsedItemT",
    bound=Union[
        span_filter_public.SpanFilterPublic,
        trace_filter_public.TraceFilterPublic,
        trace_thread_filter.TraceThreadFilter,
    ],
)
OptionalFilterParsedItemList = Optional[List[FilterParsedItemT]]

ScoreMessageT = TypeVar(
    "ScoreMessageT",
    bound=Union[messages.FeedbackScoreMessage, messages.ThreadsFeedbackScoreMessage],
)
OptionalScoreMessageList = Optional[List[ScoreMessageT]]


def datetime_to_iso8601_if_not_None(
    value: Optional[datetime.datetime],
) -> Optional[str]:
    if value is None:
        return None

    return datetime_helpers.datetime_to_iso8601(value)


def resolve_child_span_project_name(
    parent_project_name: Optional[str],
    child_project_name: Optional[str],
    show_warning: bool = True,
) -> Optional[str]:
    if parent_project_name != child_project_name:
        # if the user has specified a project name -> print warning
        if show_warning and child_project_name is not None:
            # if project name is None -> use default project name
            parent_project_name_msg = (
                parent_project_name
                if parent_project_name is not None
                else config.OPIK_PROJECT_DEFAULT_NAME
            )
            child_project_name_msg = (
                child_project_name
                if child_project_name is not None
                else config.OPIK_PROJECT_DEFAULT_NAME
            )

            LOGGER.warning(
                logging_messages.NESTED_SPAN_PROJECT_NAME_MISMATCH_WARNING_MESSAGE.format(
                    child_project_name_msg, parent_project_name_msg
                )
            )
        project_name = parent_project_name
    else:
        project_name = child_project_name

    return project_name


def add_usage_to_metadata(
    usage: Optional[Dict[str, Any]],
    metadata: Optional[Dict[str, Any]],
    create_metadata: bool = False,
) -> Optional[Dict[str, Any]]:
    if usage is None:
        return metadata

    if metadata is None and not create_metadata:
        return None

    metadata = {} if metadata is None else {**metadata}

    # Don't overwrite existing metadata.usage - it may contain original provider data
    # that should be preserved (e.g., during import from exported data)
    if "usage" in metadata:
        return metadata

    if isinstance(usage, llm_usage.OpikUsage):
        metadata["usage"] = usage.provider_usage.model_dump(exclude_none=True)
        return metadata

    metadata["usage"] = usage
    return metadata


def parse_filter_expressions(
    filter_string: Optional[str],
    parsed_item_class: Type[FilterParsedItemT],
    entity_type: Literal["traces", "threads", "spans"],
) -> OptionalFilterParsedItemList:
    """
    Parses filter expressions from a filter string using a specified class for parsed items.

    Args:
        filter_string: A string representing the filter expressions to be parsed.
        parsed_item_class: The class type to which the parsed filter expressions are mapped.
        entity_type: The entity type to determine which OQL config to use.
            Use "traces" for trace filtering, "spans" for span filtering,
            "threads" for trace thread filtering.

    Returns:
        Optional[List[T]]: A list of objects of type T created from the parsed filter
        expressions, or None if no valid expressions are found.
    """
    if filter_string is None:
        return None

    if entity_type == "spans":
        oql = opik_query_language.OpikQueryLanguage.for_spans(filter_string)
    elif entity_type == "threads":
        oql = opik_query_language.OpikQueryLanguage.for_threads(filter_string)
    else:
        oql = opik_query_language.OpikQueryLanguage.for_traces(filter_string)

    filter_expressions = oql.get_filter_expressions()

    return parse_search_expressions(
        filter_expressions, parsed_item_class=parsed_item_class
    )


def parse_search_expressions(
    filter_expressions: Optional[List[Dict[str, Any]]],
    parsed_item_class: Type[FilterParsedItemT],
) -> OptionalFilterParsedItemList:
    if filter_expressions is None:
        return None

    return [parsed_item_class(**expression) for expression in filter_expressions]


def parse_feedback_score_messages(
    scores: List[BatchFeedbackScoreDict],
    project_name: str,
    parsed_item_class: Type[ScoreMessageT],
    logger: logging.Logger,
) -> OptionalScoreMessageList:
    valid_scores = [
        score
        for score in scores
        if validation_helpers.validate_feedback_score(score, logger) is not None
    ]

    if len(valid_scores) == 0:
        return None

    score_messages = [
        parsed_item_class(
            id=score_dict["id"],
            name=score_dict["name"],
            value=score_dict["value"],
            source=constants.FEEDBACK_SCORE_SOURCE_SDK,
            project_name=score_dict.get("project_name") or project_name,
            reason=score_dict.get("reason"),
            category_name=score_dict.get("category_name"),
        )
        for score_dict in valid_scores
    ]

    return score_messages



from typing import Dict, Optional, Any, List

from rich import align
from rich.console import Console
from rich.table import Table
from rich.text import Text

import opik.config as config
from opik.config import OpikConfig

DEFAULT_KEY_COLOR = "green"
DEFAULT_VALUE_COLOR = "blue"
DEFAULT_ERROR_COLOR = "red"

console = Console()


def make_key_text(key: str) -> Text:
    return Text(key, style=DEFAULT_KEY_COLOR)


def make_value_text(value: str) -> Text:
    return Text(value, style=DEFAULT_VALUE_COLOR)


def print_header(text: str) -> None:
    header_text = f"*** {text.upper()} ***"
    header_text = Text(header_text, style="bold magenta")
    header_text = align.Align.left(header_text)

    console.print()
    console.print(header_text)


def print_installed_packages(packages: Dict[str, str]) -> None:
    for name, version in sorted(packages.items()):
        name = make_key_text(name)
        version = make_value_text(version)

        console.print(name, "==", version, sep="")


def print_versions(python_version: str, opik_version: str) -> None:
    python_version_label = make_key_text("Python version:")
    python_version = make_value_text(python_version)
    opik_version_label = make_key_text("Opik version:")
    opik_version = make_value_text(opik_version)

    console.print(python_version_label, python_version)
    console.print(opik_version_label, opik_version)


def print_config_file_details(config: OpikConfig) -> None:
    file_path_label = make_key_text("Config file path:")
    file_path = make_value_text(str(config.config_file_fullpath))

    is_exists_label = make_key_text("Config file exists:")
    is_exists = make_value_text("yes" if config.config_file_exists else "no")

    console.print(file_path_label, file_path)
    console.print(is_exists_label, is_exists)


def print_current_config(config: config.OpikConfig) -> None:
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Setting", style=DEFAULT_KEY_COLOR)
    table.add_column("Value", style=DEFAULT_VALUE_COLOR)

    current_config_values = config.as_dict(mask_api_key=True)
    for key, value in sorted(current_config_values.items()):
        if key != "sentry_dsn":
            table.add_row(key, str(value))

    console.print(table)


def print_config_scan_results(
    misconfiguration_detected: bool, error_message: Optional[str]
) -> None:
    is_misconfigured_text = Text(
        "found" if misconfiguration_detected else "not found",
        style=DEFAULT_ERROR_COLOR if misconfiguration_detected else DEFAULT_VALUE_COLOR,
    )
    issues_found_label = make_key_text("Configuration issues:")

    console.print(issues_found_label, is_misconfigured_text)

    if not misconfiguration_detected:
        return

    err_msg = Text(error_message, style=DEFAULT_ERROR_COLOR)
    console.print(err_msg)


def print_backend_workspace_availability(
    is_available: bool,
    err_msg: Optional[str],
) -> None:
    is_available_text = Text(
        "yes" if is_available else "no",
        style=DEFAULT_VALUE_COLOR if is_available else DEFAULT_ERROR_COLOR,
    )
    is_available_label = make_key_text("Backend workspace available:")

    console.print(is_available_label, is_available_text)

    if err_msg:
        print_error_message(err_msg)


def print_error_message(err_msg: str) -> None:
    err_msg = Text(err_msg, style=DEFAULT_ERROR_COLOR)
    console.print(err_msg)


def print_opik_permissions_url(opik_url: str) -> None:
    opik_url_label = make_key_text("OPIK url:")
    opik_url = make_value_text(opik_url)
    console.print(opik_url_label, opik_url)


def print_user_permissions(user_permissions: Dict[str, Any]) -> None:
    user_label = make_key_text("User:")
    user_name = make_value_text(user_permissions["user_name"])
    workspace_label = make_key_text("Workspace:")
    workspace_name = make_value_text(user_permissions["workspace_name"])
    console.print(user_label, user_name)
    console.print(workspace_label, workspace_name)

    # print permissions
    permissions: List[Dict[str, str]] = user_permissions["permissions"]

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Permission", style=DEFAULT_KEY_COLOR)
    table.add_column("Available", style=DEFAULT_VALUE_COLOR)

    for permission in permissions:
        value = permission["permission_value"] == "true"
        table.add_row(permission["permission_name"], str(value))

    console.print(table)



import logging
from typing import Optional

from opik import semantic_version
from opik.api_objects import opik_client
from opik.types import DistributedTraceHeadersDict
from . import llm_response_wrapper
from . import litellm_wrappers
from .adk_otel_tracer import opik_adk_otel_tracer

import google.adk.models
from google.adk.models import lite_llm
from google.adk import telemetry as adk_telemetry
from google.adk.agents import base_agent

LOGGER = logging.getLogger(__name__)


def patch_adk(
    opik_client: opik_client.Opik,
    distributed_headers: Optional[DistributedTraceHeadersDict] = None,
) -> None:
    _patch_adk_lite_llm()
    if semantic_version.SemanticVersion.parse(google.adk.__version__) >= "1.3.0":  # type: ignore
        _patch_adk_opentelemetry_tracers(
            opik_client, distributed_headers=distributed_headers
        )


def _patch_adk_lite_llm() -> None:
    old_function = google.adk.models.LlmResponse.create
    if not hasattr(old_function, "opik_patched"):
        create_wrapper = llm_response_wrapper.LlmResponseCreateWrapper(old_function)
        google.adk.models.LlmResponse.create = create_wrapper
        google.adk.models.LlmResponse.create.opik_patched = True  # type: ignore
        LOGGER.debug("Patched LlmResponse.create")

    if hasattr(lite_llm, "LiteLLMClient") and hasattr(
        lite_llm.LiteLLMClient, "acompletion"
    ):
        if not hasattr(lite_llm.LiteLLMClient.acompletion, "opik_patched"):
            lite_llm.LiteLLMClient.acompletion = (
                litellm_wrappers.litellm_client_acompletion_decorator(
                    lite_llm.LiteLLMClient.acompletion
                )
            )
            lite_llm.LiteLLMClient.acompletion.opik_patched = True  # type: ignore
            LOGGER.debug("Patched LiteLLMClient.acompletion")

    if hasattr(lite_llm, "_model_response_to_generate_content_response"):
        if not hasattr(
            lite_llm._model_response_to_generate_content_response, "opik_patched"
        ):
            lite_llm._model_response_to_generate_content_response = (
                litellm_wrappers.generate_content_response_decorator(
                    lite_llm._model_response_to_generate_content_response
                )
            )
            lite_llm._model_response_to_generate_content_response.opik_patched = True  # type: ignore
            LOGGER.debug("Patched _model_response_to_generate_content_response")


def _patch_adk_opentelemetry_tracers(
    opik_client: opik_client.Opik,
    distributed_headers: Optional[DistributedTraceHeadersDict],
) -> None:
    no_op_opik_tracer = opik_adk_otel_tracer.OpikADKOtelTracer(
        opik_client, distributed_headers=distributed_headers
    )

    adk_telemetry.tracer.start_as_current_span = no_op_opik_tracer.start_as_current_span
    adk_telemetry.tracer.start_span = no_op_opik_tracer.start_span

    base_agent.tracer.start_as_current_span = no_op_opik_tracer.start_as_current_span
    base_agent.tracer.start_span = no_op_opik_tracer.start_span
    LOGGER.debug("Patched ADK tracers")



from typing import List, Optional, Any
from enum import Enum

import google.adk.agents
import google.adk.tools.agent_tool
import inspect


class GraphNodeType(str, Enum):
    SEQUENTIAL_AGENT = "sequential_agent"
    LOOP_AGENT = "loop_agent"
    PARALLEL_AGENT = "parallel_agent"
    LLM_AGENT = "llm_agent"
    TOOL = "tool"


class AgentNode:
    def __init__(self, agent: google.adk.agents.BaseAgent) -> None:
        self.agent = agent
        self.name = _to_mermaid_compatible_name(agent.name)
        self.subagent_nodes: List[AgentNode] = []
        self.type = _determine_agent_type(agent)
        self.tools: List[ToolNode] = []

    def add_child(self, child: "AgentNode") -> None:
        self.subagent_nodes.append(child)

    def add_tool(self, tool: "ToolNode") -> None:
        self.tools.append(tool)

    @property
    def first_child_name(self) -> Optional[str]:
        return self.subagent_nodes[0].name if self.subagent_nodes else None

    @property
    def last_child_name(self) -> Optional[str]:
        return self.subagent_nodes[-1].name if self.subagent_nodes else None


class ToolNode:
    def __init__(self, tool: Any, name: str) -> None:
        self.tool = tool
        self.name = name
        self.agent: Optional[AgentNode] = None
        self.type = GraphNodeType.TOOL

    def associate_agent_node(self, agent: AgentNode) -> None:
        """Used to link a tool node with agent node if a tool is an agent"""
        self.agent = agent


def _determine_agent_type(agent: google.adk.agents.BaseAgent) -> GraphNodeType:
    if isinstance(agent, google.adk.agents.SequentialAgent):
        return GraphNodeType.SEQUENTIAL_AGENT
    elif isinstance(agent, google.adk.agents.LoopAgent):
        return GraphNodeType.LOOP_AGENT
    elif isinstance(agent, google.adk.agents.ParallelAgent):
        return GraphNodeType.PARALLEL_AGENT
    else:
        return GraphNodeType.LLM_AGENT


def _to_mermaid_compatible_name(name: str) -> str:
    return name.replace(" ", "_")


def build_nodes_tree(agent: google.adk.agents.BaseAgent) -> AgentNode:
    node = AgentNode(agent)

    sub_agents = getattr(agent, "sub_agents", []) or []

    for sub_agent in sub_agents:
        child_node = build_nodes_tree(sub_agent)
        node.add_child(child_node)

    tools = getattr(agent, "tools", [])
    for tool in tools:
        tool_node = ToolNode(tool, name=_extract_tool_name(tool))
        if isinstance(tool, google.adk.tools.agent_tool.AgentTool):
            tool_agent_node = build_nodes_tree(tool.agent)
            tool_node.associate_agent_node(tool_agent_node)

        node.add_tool(tool_node)

    return node


def _extract_tool_name(tool: Any) -> str:
    if isinstance(tool, google.adk.tools.agent_tool.AgentTool):
        return _to_mermaid_compatible_name(f"AgentTool:{tool.name}")

    if inspect.isfunction(tool):
        return tool.__name__

    if hasattr(tool, "name") and isinstance(tool.name, str):
        return _to_mermaid_compatible_name(tool.name)

    if hasattr(tool, "__class__"):
        return tool.__class__.__name__

    return _to_mermaid_compatible_name(str(tool))



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .alert_public_alert_type import AlertPublicAlertType
from .alert_trigger_public import AlertTriggerPublic
from .webhook_public import WebhookPublic


class AlertPublic(UniversalBaseModel):
    id: typing.Optional[str] = None
    name: typing.Optional[str] = None
    enabled: typing.Optional[bool] = None
    alert_type: typing.Optional[AlertPublicAlertType] = None
    metadata: typing.Optional[typing.Dict[str, str]] = None
    webhook: WebhookPublic
    triggers: typing.Optional[typing.List[AlertTriggerPublic]] = None
    created_at: typing.Optional[dt.datetime] = None
    created_by: typing.Optional[str] = None
    last_updated_at: typing.Optional[dt.datetime] = None
    last_updated_by: typing.Optional[str] = None
    project_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from typing import (
    Optional,
    NamedTuple,
)

import opik.context_storage as context_storage
import opik.datetime_helpers as datetime_helpers
from opik.api_objects import helpers, span, trace
from opik.types import DistributedTraceHeadersDict, TraceSource

from . import arguments_helpers


class SpanCreationResult(NamedTuple):
    """
    Represents the result of a span creation process.

    This class encapsulates the data resulting from the creation of a new
    span, including trace information and span-specific details.

    Attributes:
        trace_data: Trace-related data associated
            with the span if a new trace was created. Can be None if no new trace was created.
        span_data : Data specific to the created span, containing
            information such as span identifiers and timestamps.
        should_process_span_data: A boolean indicating whether created span data should be further processed
            after it was created (saved, logged, etc.).
    """

    trace_data: Optional[trace.TraceData]
    span_data: span.SpanData
    should_process_span_data: bool


def create_span_respecting_context(
    start_span_arguments: arguments_helpers.StartSpanParameters,
    distributed_trace_headers: Optional[DistributedTraceHeadersDict],
    opik_context_storage: Optional[context_storage.OpikContextStorage] = None,
    should_create_duplicate_root_span: bool = True,
    preset_trace_id: Optional[str] = None,
    source: Optional[TraceSource] = None,
) -> SpanCreationResult:
    """
    Handles different span creation flows.
    """

    if opik_context_storage is None:
        opik_context_storage = context_storage.get_current_context_instance()

    if distributed_trace_headers:
        span_data = arguments_helpers.create_span_data(
            start_span_arguments=start_span_arguments,
            parent_span_id=distributed_trace_headers["opik_parent_span_id"],
            trace_id=distributed_trace_headers["opik_trace_id"],
            source=source,
        )

        return SpanCreationResult(None, span_data, should_process_span_data=True)

    current_span_data = opik_context_storage.top_span_data()
    current_trace_data = opik_context_storage.get_trace_data()

    if current_span_data is not None:
        # There is already at least one span in the current context - attach a new span to it.
        #
        # NOTE: We can have a situation when span data is in context, but there is no trace data
        # because we are in a distributed environment and trace data was created in another thread.
        # See: https://github.com/comet-ml/opik/pull/2244
        if current_trace_data is None:
            show_warning = False
        else:
            show_warning = current_trace_data.created_by != "evaluation"

        project_name = helpers.resolve_child_span_project_name(
            parent_project_name=current_span_data.project_name,
            child_project_name=start_span_arguments.project_name,
            show_warning=show_warning,
        )

        start_span_arguments.project_name = project_name

        span_data = arguments_helpers.create_span_data(
            start_span_arguments=start_span_arguments,
            parent_span_id=current_span_data.id,
            trace_id=current_span_data.trace_id,
            source=source if source is not None else current_span_data.source,
        )

        return SpanCreationResult(None, span_data, should_process_span_data=True)

    if current_trace_data is not None and current_span_data is None:
        # By default, we expect trace to be created with a span.
        # But there can be cases when trace was created and added
        # to context manually (not via decorator).
        # In that case the decorator should just create a span for the existing trace.

        project_name = helpers.resolve_child_span_project_name(
            parent_project_name=current_trace_data.project_name,
            child_project_name=start_span_arguments.project_name,
            show_warning=current_trace_data.created_by != "evaluation",
        )

        start_span_arguments.project_name = project_name

        span_data = arguments_helpers.create_span_data(
            start_span_arguments=start_span_arguments,
            parent_span_id=None,
            trace_id=current_trace_data.id,
            source=source if source is not None else current_trace_data.source,
        )

        return SpanCreationResult(None, span_data, should_process_span_data=True)

    if current_span_data is None and current_trace_data is None:
        # Create a trace and root span because it is
        # the first decorated function run in the current context.
        current_trace_data = trace.TraceData(
            id=preset_trace_id or helpers.generate_id(),
            start_time=datetime_helpers.local_timestamp(),
            name=start_span_arguments.name,
            input=start_span_arguments.input,
            metadata=start_span_arguments.metadata,
            tags=start_span_arguments.tags,
            project_name=start_span_arguments.project_name,
            thread_id=start_span_arguments.thread_id,
            source=source if source is not None else "sdk",
        )

        current_span_data = arguments_helpers.create_span_data(
            start_span_arguments=start_span_arguments,
            parent_span_id=None,
            trace_id=current_trace_data.id,
            source=source,
        )

    return SpanCreationResult(
        current_trace_data,
        current_span_data,
        should_process_span_data=should_create_duplicate_root_span,
    )



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.service_toggles_config import ServiceTogglesConfig
from .raw_client import AsyncRawServiceTogglesClient, RawServiceTogglesClient


class ServiceTogglesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawServiceTogglesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawServiceTogglesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawServiceTogglesClient
        """
        return self._raw_client

    def get_service_toggles(self, *, request_options: typing.Optional[RequestOptions] = None) -> ServiceTogglesConfig:
        """
        Get Service Toggles

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ServiceTogglesConfig
            Service Toggles

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.service_toggles.get_service_toggles()
        """
        _response = self._raw_client.get_service_toggles(request_options=request_options)
        return _response.data


class AsyncServiceTogglesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawServiceTogglesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawServiceTogglesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawServiceTogglesClient
        """
        return self._raw_client

    async def get_service_toggles(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ServiceTogglesConfig:
        """
        Get Service Toggles

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ServiceTogglesConfig
            Service Toggles

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.service_toggles.get_service_toggles()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_service_toggles(request_options=request_options)
        return _response.data



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.welcome_wizard_tracking import WelcomeWizardTracking
from .raw_client import AsyncRawWelcomeWizardClient, RawWelcomeWizardClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class WelcomeWizardClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawWelcomeWizardClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawWelcomeWizardClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawWelcomeWizardClient
        """
        return self._raw_client

    def get_welcome_wizard_status(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WelcomeWizardTracking:
        """
        Get welcome wizard tracking status for the current workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WelcomeWizardTracking
            Welcome wizard tracking status

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.welcome_wizard.get_welcome_wizard_status()
        """
        _response = self._raw_client.get_welcome_wizard_status(request_options=request_options)
        return _response.data

    def submit_welcome_wizard(
        self,
        *,
        role: typing.Optional[str] = OMIT,
        integrations: typing.Optional[typing.Sequence[str]] = OMIT,
        email: typing.Optional[str] = OMIT,
        join_beta_program: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Submit welcome wizard with user information

        Parameters
        ----------
        role : typing.Optional[str]
            Optional user role

        integrations : typing.Optional[typing.Sequence[str]]
            List of integrations the user selected

        email : typing.Optional[str]
            Optional user email

        join_beta_program : typing.Optional[bool]
            Whether user wants to join beta programs

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.welcome_wizard.submit_welcome_wizard()
        """
        _response = self._raw_client.submit_welcome_wizard(
            role=role,
            integrations=integrations,
            email=email,
            join_beta_program=join_beta_program,
            request_options=request_options,
        )
        return _response.data


class AsyncWelcomeWizardClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawWelcomeWizardClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawWelcomeWizardClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawWelcomeWizardClient
        """
        return self._raw_client

    async def get_welcome_wizard_status(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WelcomeWizardTracking:
        """
        Get welcome wizard tracking status for the current workspace

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WelcomeWizardTracking
            Welcome wizard tracking status

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.welcome_wizard.get_welcome_wizard_status()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_welcome_wizard_status(request_options=request_options)
        return _response.data

    async def submit_welcome_wizard(
        self,
        *,
        role: typing.Optional[str] = OMIT,
        integrations: typing.Optional[typing.Sequence[str]] = OMIT,
        email: typing.Optional[str] = OMIT,
        join_beta_program: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Submit welcome wizard with user information

        Parameters
        ----------
        role : typing.Optional[str]
            Optional user role

        integrations : typing.Optional[typing.Sequence[str]]
            List of integrations the user selected

        email : typing.Optional[str]
            Optional user email

        join_beta_program : typing.Optional[bool]
            Whether user wants to join beta programs

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.welcome_wizard.submit_welcome_wizard()
        asyncio.run(main())
        """
        _response = await self._raw_client.submit_welcome_wizard(
            role=role,
            integrations=integrations,
            email=email,
            join_beta_program=join_beta_program,
            request_options=request_options,
        )
        return _response.data
